package tokenize_test

import (
	"fmt"
	"strings"
	"testing"

	"git.makaay.nl/mauricem/go-parsekit/tokenize"
)

func ExampleNewAPI() {
	tokenize.NewAPI("The input that the API will handle")

	// Output:
}

// func ExampleAPI_NextRune() {
// 	api := tokenize.NewAPI("The input that the API will handle")
// 	r, err := api.NextRune()
// 	fmt.Printf("Rune read from input; %c\n", r)
// 	fmt.Printf("The error: %v\n", err)
// 	fmt.Printf("API results: %q\n", api.Output.String())

// 	// Output:
// 	// Rune read from input; T
// 	// The error: <nil>
// 	// API results: ""
// }

func ExampleAPI_PeekRune() {
	api := tokenize.NewAPI("The input that the API will handle")

	r1, _, err := api.Input.PeekRune(19) // 'A'
	r2, _, err := api.Input.PeekRune(20) // 'P'
	r3, _, err := api.Input.PeekRune(21) // 'I'
	_, _, err = api.Input.PeekRune(100)  // EOF

	fmt.Printf("%c%c%c %s\n", r1, r2, r3, err)

	// Output:
	// API EOF
}

func ExampleAPI_AcceptRune() {
	api := tokenize.NewAPI("The input that the API will handle")

	// Reads 'T' and accepts it to the API results.
	r, _, _ := api.Input.PeekRune(0)
	api.Input.AcceptRune(r)

	// Reads 'h' and accepts it to the API results.
	r, _, _ = api.Input.PeekRune(0)
	api.Input.AcceptRune(r)

	// Reads 'e', but does not accept it to the API results.
	r, _, _ = api.Input.PeekRune(0)

	fmt.Printf("API results: %q\n", api.Output.String())

	// Output:
	// API results: "Th"
}

func ExampleAPI_AcceptRunes() {
	api := tokenize.NewAPI("The input that the API will handle")

	// Peeks at the first two runes 'T' and 'h'.
	r0, _, _ := api.Input.PeekRune(0)
	r1, _, _ := api.Input.PeekRune(1)

	// Peeks at the third rune 'e'.
	api.Input.PeekRune(2)

	// Accepts only 'T' and 'h' into the API results.
	api.Input.AcceptRunes(r0, r1)

	fmt.Printf("API results: %q\n", api.Output.String())

	// Output:
	// API results: "Th"
}

func ExampleAPI_SkipRune() {
	api := tokenize.NewAPI("The input that the API will handle")

	for {
		r, _, err := api.Input.PeekRune(0)

		// EOF reached.
		if err != nil {
			break
		}

		// Only accept runes that are vowels.
		if strings.ContainsRune("aeiouAEIOU", r) {
			api.Input.AcceptRune(r)
		} else {
			api.Input.SkipRune(r)
		}
	}

	fmt.Printf("API results: %q\n", api.Output.String())

	// Output:
	// API results: "eiuaeAIiae"
}

func ExampleAPI_modifyingResults() {
	api := tokenize.NewAPI("")

	api.Output.AddString("Some runes")
	api.Output.AddRunes(' ', 'a', 'd', 'd', 'e', 'd')
	api.Output.AddRunes(' ', 'i', 'n', ' ')
	api.Output.AddString("various ways")
	fmt.Printf("API result first 10 runes: %q\n", api.Output.Runes()[0:10])
	fmt.Printf("API result runes as string: %q\n", api.Output.String())

	api.Output.SetString("new ")
	api.Output.AddString("set ")
	api.Output.AddString("of ")
	api.Output.AddRunes('r', 'u', 'n', 'e', 's')
	fmt.Printf("API result runes as string: %q\n", api.Output.String())
	fmt.Printf("API result runes: %q\n", api.Output.Runes())
	fmt.Printf("API third rune: %q\n", api.Output.Rune(2))

	api.Output.AddTokens(tokenize.Token{
		Type:  42,
		Value: "towel"})
	api.Output.AddTokens(tokenize.Token{
		Type:  73,
		Value: "Zaphod"})
	fmt.Printf("API result tokens: %v\n", api.Output.Tokens())
	fmt.Printf("API second result token: %v\n", api.Output.Token(1))

	// Output:
	// API result first 10 runes: ['S' 'o' 'm' 'e' ' ' 'r' 'u' 'n' 'e' 's']
	// API result runes as string: "Some runes added in various ways"
	// API result runes as string: "new set of runes"
	// API result runes: ['n' 'e' 'w' ' ' 's' 'e' 't' ' ' 'o' 'f' ' ' 'r' 'u' 'n' 'e' 's']
	// API third rune: 'w'
	// API result tokens: [42("towel") 73("Zaphod")]
	// API second result token: 73("Zaphod")
}

func ExampleAPI_Reset() {
	api := tokenize.NewAPI("Very important input!")

	r, _, _ := api.Input.PeekRune(0) // read 'V'
	api.Input.AcceptRune(r)
	r, _, _ = api.Input.PeekRune(0) // read 'e'
	api.Input.AcceptRune(r)
	fmt.Printf("API results: %q at %s\n", api.Output.String(), api.Input.Cursor())

	// Reset clears the results.
	api.Input.Reset()
	fmt.Printf("API results: %q at %s\n", api.Output.String(), api.Input.Cursor())

	// So then doing the same read operations, the same data are read.
	r, _, _ = api.Input.PeekRune(0) // read 'V'
	api.Input.AcceptRune(r)
	r, _, _ = api.Input.PeekRune(0) // read 'e'
	api.Input.AcceptRune(r)
	fmt.Printf("API results: %q at %s\n", api.Output.String(), api.Input.Cursor())

	// Output:
	// API results: "Ve" at line 1, column 3
	// API results: "" at start of file
	// API results: "Ve" at line 1, column 3
}

func ExampleAPI_Fork() {
	// This custom Handler checks for input 'a', 'b' or 'c'.
	abcHandler := func(t *tokenize.API) bool {
		a := tokenize.A
		for _, r := range []rune{'a', 'b', 'c'} {
			child := t.Fork() // fork, so we won't change parent t
			if a.Rune(r)(t) {
				t.Merge(child)   // accept results into parent of child
				t.Dispose(child) // return to the parent level
				return true      // and report a successful match
			}
			t.Dispose(child) // return to the parent level
		}
		// If we get here, then no match was found. Return false to communicate
		// this to the caller.
		return false
	}

	// Note: a custom Handler is normally not what you need.
	// You can make use of the parser/combinator tooling to make the
	// implementation a lot simpler and to take care of forking at
	// the appropriate places. The handler from above can be replaced with:
	simpler := tokenize.A.RuneRange('a', 'c')

	result, err := tokenize.New(abcHandler)("another test")
	fmt.Println(result, err)
	result, err = tokenize.New(simpler)("curious")
	fmt.Println(result, err)
	result, err = tokenize.New(abcHandler)("bang on!")
	fmt.Println(result, err)
	result, err = tokenize.New(abcHandler)("not a match")
	fmt.Println(result, err)

	// Output:
	// a <nil>
	// c <nil>
	// b <nil>
	// <nil> mismatch at start of file
}

func ExampleAPI_Merge() {
	tokenHandler := func(t *tokenize.API) bool {
		child1 := t.Fork()
		r0, _, _ := t.Input.PeekRune(0) // reads 'H'
		r1, _, _ := t.Input.PeekRune(1) // reads 'i'
		t.Input.AcceptRunes(r0, r1)     // these runes are accepted in the API results for child1

		child2 := t.Fork()
		r0, _, _ = t.Input.PeekRune(0) // reads ' '
		r1, _, _ = t.Input.PeekRune(1) // reads 'm'
		t.Input.AcceptRunes(r0, r1)    // these runes are accepted in the API results for child2
		t.Dispose(child2)              // but they are not merged and thefore not used by child1

		t.Merge(child1)   // We merge child1, which has read 'H' and 'i' only.
		t.Dispose(child1) // and clean up child1 to return to the parent
		return true
	}

	result, _ := tokenize.New(tokenHandler)("Hi mister X!")
	fmt.Println(result.String())

	// Output:
	// Hi
}

func TestMultipleLevelsOfForksAndMerges(t *testing.T) {
	api := tokenize.NewAPI("abcdefghijklmnopqrstuvwxyz")

	// Fork a few levels.
	child1 := api.Fork()
	child2 := api.Fork()
	child3 := api.Fork()
	child4 := api.Fork()

	// Read a rune 'a' from child4.
	r, _, _ := api.Input.PeekRune(0)
	AssertEqual(t, 'a', r, "child4 rune 1")
	api.Input.AcceptRune(r)
	AssertEqual(t, "a", api.Output.String(), "child4 runes after rune 1")

	// Read another rune 'b' from child4.
	r, _, _ = api.Input.PeekRune(0)
	AssertEqual(t, 'b', r, "child4 rune 2")
	api.Input.AcceptRune(r)
	AssertEqual(t, "ab", api.Output.String(), "child4 runes after rune 2")

	// Merge "ab" from child4 to child3.
	api.Merge(child4)
	AssertEqual(t, "", api.Output.String(), "child4 runes after first merge")

	// Read some more from child4.
	r, _, _ = api.Input.PeekRune(0)
	AssertEqual(t, 'c', r, "child4 rune 3")
	api.Input.AcceptRune(r)
	AssertEqual(t, "c", api.Output.String(), "child4 runes after rune 1")
	AssertEqual(t, "line 1, column 4", api.Input.Cursor(), "cursor child4 rune 3")

	// Merge "c" from child4 to child3.
	api.Merge(child4)

	// And dispose of child4, making child3 the active stack level.
	api.Dispose(child4)

	// Child3 should now have the compbined results "abc" from child4's work.
	AssertEqual(t, "abc", api.Output.String(), "child3 after merge of child4")
	AssertEqual(t, "line 1, column 4", api.Input.Cursor(), "cursor child3 rune 3, after merge of child4")

	// Now read some data from child3.
	r, _, _ = api.Input.PeekRune(0)
	AssertEqual(t, 'd', r, "child3 rune 5")
	api.Input.AcceptRune(r)

	r, _, _ = api.Input.PeekRune(0)
	AssertEqual(t, 'e', r, "child3 rune 5")
	api.Input.AcceptRune(r)

	r, _, _ = api.Input.PeekRune(0)
	AssertEqual(t, 'f', r, "child3 rune 5")
	api.Input.AcceptRune(r)

	AssertEqual(t, "abcdef", api.Output.String(), "child3 total result after rune 6")

	// Temporarily go some new forks from here, but don't use their outcome.
	child3sub1 := api.Fork()
	r, _, _ = api.Input.PeekRune(0)
	api.Input.AcceptRune(r)
	r, _, _ = api.Input.PeekRune(0)
	api.Input.AcceptRune(r)
	child3sub2 := api.Fork()
	r, _, _ = api.Input.PeekRune(0)
	api.Input.AcceptRune(r)
	api.Merge(child3sub2)   // do merge sub2 down to sub1
	api.Dispose(child3sub2) // and dispose of sub2
	api.Dispose(child3sub1) // but dispose of sub1 without merging

	// Instead merge the results from before this forking segway from child3 to child2
	// and dispose of it.
	api.Merge(child3)
	api.Dispose(child3)

	AssertEqual(t, "abcdef", api.Output.String(), "child2 total result after merge of child3")
	AssertEqual(t, "line 1, column 7", api.Input.Cursor(), "cursor child2 after merge child3")

	// Merge child2 to child1 and dispose of it.
	api.Merge(child2)
	api.Dispose(child2)

	// Merge child1 a few times to the top level api.
	api.Merge(child1)
	api.Merge(child1)
	api.Merge(child1)
	api.Merge(child1)

	// And dispose of it.
	api.Dispose(child1)

	// Read some data from the top level api.
	r, _, _ = api.Input.PeekRune(0)
	api.Input.AcceptRune(r)

	AssertEqual(t, "abcdefg", api.Output.String(), "api string end result")
	AssertEqual(t, "line 1, column 8", api.Input.Cursor(), "api cursor end result")
}

func TestClearData(t *testing.T) {
	api := tokenize.NewAPI("Laphroaig")
	r, _, _ := api.Input.PeekRune(0) // Read 'L'
	api.Input.AcceptRune(r)          // Add to runes
	r, _, _ = api.Input.PeekRune(0)  // Read 'a'
	api.Input.AcceptRune(r)          // Add to runes
	api.Output.ClearData()           // Clear the runes, giving us a fresh start.
	r, _, _ = api.Input.PeekRune(0)  // Read 'p'
	api.Input.AcceptRune(r)          // Add to runes
	r, _, _ = api.Input.PeekRune(0)  // Read 'r'
	api.Input.AcceptRune(r)          // Add to runes

	AssertEqual(t, "ph", api.Output.String(), "api string end result")
}

func TestMergeScenariosForTokens(t *testing.T) {
	api := tokenize.NewAPI("")

	token1 := tokenize.Token{Value: 1}
	token2 := tokenize.Token{Value: 2}
	token3 := tokenize.Token{Value: 3}
	token4 := tokenize.Token{Value: 4}

	api.Output.SetTokens(token1)
	tokens := api.Output.Tokens()
	AssertEqual(t, 1, len(tokens), "Tokens 1")

	child := api.Fork()

	tokens = api.Output.Tokens()
	AssertEqual(t, 0, len(tokens), "Tokens 2")

	api.Output.AddTokens(token2)

	api.Merge(child)
	api.Dispose(child)

	tokens = api.Output.Tokens()
	AssertEqual(t, 2, len(tokens), "Tokens 3")

	child = api.Fork()
	api.Output.AddTokens(token3)
	api.Input.Reset()
	api.Output.AddTokens(token4)

	api.Merge(child)
	api.Dispose(child)

	tokens = api.Output.Tokens()
	AssertEqual(t, 3, len(tokens), "Tokens 4")
	AssertEqual(t, 1, api.Output.TokenValue(0).(int), "Tokens 4, value 0")
	AssertEqual(t, 2, api.Output.TokenValue(1).(int), "Tokens 4, value 1")
	AssertEqual(t, 4, api.Output.TokenValue(2).(int), "Tokens 4, value 2")
}
